// Time-stamp: "26.04.21 12:06 meinTollesModul.hpp klaus.wachtler@esolutions.de"
//
// meinTollesModul liefert irgendwelche Daten.
// Nur zum Spielen wird ein Thread gestartet, der alle paar Sekunden
// die Uhrzeit holt.
//
// Ein Nutzer der Daten kann sich Änderungen abonnieren.
// Wenn sich die Uhrzeit ändert, werden alle Nutzer benachrichtigt.


#include <time.h>

#include <pthread.h>
#include <unistd.h>

#include <stdexcept>
#include <set>


class MeinTollesModul
{
public:

  // Konstruktor: initialisiert das MeinTollesModul-Objekt und startet
  // den Thread
  MeinTollesModul()
    : m_time( time( NULL ) )
  {
    pthread_attr_t    attr;
    if( pthread_attr_init( &attr )!=0 )
    {
      throw ThreadError( "pthread_attr_init() failed." );
    }
    if( pthread_create( &m_threadId, &attr, timeThread, this )!=0 )
    {
      throw ThreadError( "pthread_create() failed." );
    }
  }

  // So etwas wird evtl. als Fehler geworfen:
  class ThreadError : public std::runtime_error
  {
  public:
    ThreadError( const std::string &_what = "ThreadError" )
      : std::runtime_error( std::string( "Error on creating thread\n" ) + _what )
    {}
  };

  // Typ für eine callback-Funktion (mit der sich ein Nutzer
  // registrieren kann)
  // Also ein Zeiger auf eine Funktion, die bei einer geänderten
  // Uhrzeit aufgerufen wird.
  typedef  void (*callbackTimeFkt_t)( time_t t );

  // trägt einen weiteren Benutzer ein:
  void subscribe( callbackTimeFkt_t callbackFkt )
  {
    // seine callback-Funktion im set ablegen:
    m_Nutzer.insert(  callbackFkt );
    // Weil der neue Nutzer die aktuellen Daten noch nicht kennt,
    // schicken wir sie ihm gleich:
    callbackFkt( m_time );
  }

  // entfernt einen Benutzer:
  void unsubscribe( callbackTimeFkt_t callbackFkt )
  {
    //std::set<callbackTimeFkt_t>::iterator    gefunden;
    m_Nutzer.erase( callbackFkt );
  }

private:

  // der Hintergrundthread: aktiviert gelegentlich triggerTimeEvent()
  static   void *timeThread( void *pMeinTollesModul )
  {
    while( "Putin ist ein Gangster" )
    {
      ((MeinTollesModul*)pMeinTollesModul)->triggerTimeEvent();
      sleep( 1 );
    }

    return NULL;
  }

  // wird gelegentlich aufgerufen, holt die aktuelle Uhrzeit und
  // benachrichtigt alle Nutzer
  void triggerTimeEvent()
  {
    time_t now = time( NULL );
    if( now!=m_time )
    {
      // neue Zeit merken:
      m_time = now;

      // Alle angemeldeten Nutzer benachrichtigen (für jeden seine
      // callback-Funktion aufrufen):
      for( std::set<callbackTimeFkt_t>::const_iterator iNutzer=m_Nutzer.begin();
           iNutzer!=m_Nutzer.end();
           ++iNutzer
           )
      {
        (*iNutzer)( m_time );
      }
    }
  }

  time_t                            m_time;
  pthread_t                         m_threadId;

  std::set<callbackTimeFkt_t>       m_Nutzer;
};


